/*  file: schedule.c
 *
 *  kehpager, Charset aware pager, Kari E. Hurtta
 *
 *  Copyright (c) 1994 Kari E. Hurtta
 *
 *  Redistribution and use in source and binary forms are permitted
 *  provided that the above copyright notice and this paragraph are
 *  duplicated in all such forms. This software is provided 'as is'
 *  and without any warranty. 
 */

#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
/* #include <signal.h> terminal.h includes this */

#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#if defined(SUNOS4) || defined(SUNOS5)
#include <sys/filio.h>
#else
#include <sys/ioctl.h>
#endif
#include <unistd.h>

#include <time.h>


#include <errno.h>
extern int errno;

#include "kehpager.h"
#include "memory.h"
#include "charset.h"

#include "schedule.h"

#include "terminal.h"

#include "control.h"
#include "esc.h"
#include "window.h"
#include "keyboard.h"

int cancel = 0;

void real_schedule(int *FP, int *EOF_IN_FILE, int *NOT_SELECTABLE,
		  void (*file_handler)(void),
		  int (*timeout_callback)(long *value),
		  void (*timeout_handler)(int key),
		  int *ready, int (* callback_before)(void),
		  void (* callback_after)(void),
		  int call_parser
		  ) {
#ifdef _HPUX_SOURCE
  int readfs,writefs;
#else
  fd_set readfs,writefs;
#endif
  struct timeval *timeout,Timeout;
  long atime;
  int res;
  int i;

  int tmp_ready=0;
  if (!ready)
    ready = &tmp_ready;

  while (!(*ready) || (!tmp_ready && !quitflag)) {
    int skip = 0, was_input = 0;
    int biggest = terminal;
    int have_file;
    int timeout_key = 0;

    tmp_ready = 1;
    
    skip = (* callback_before)();

    /* check pointers */
    have_file = FP && EOF_IN_FILE && NOT_SELECTABLE;

#ifdef _HPUX_SOURCE
    readfs = 1 << terminal;    
    writefs = 0;
#else
    FD_ZERO(&readfs);
    FD_SET(terminal,&readfs);
    FD_ZERO(&writefs);
#endif

    if (need_line_flush() || need_flush_buffer()) {
#ifdef _HPUX_SOURCE
      writefs |= 1 << terminal;
#else
      FD_SET(terminal,&writefs);
#endif
    }

    
    if (have_file && -1 != (*FP) && !(*EOF_IN_FILE)) {
      if (!(*NOT_SELECTABLE)) {
#ifdef _HPUX_SOURCE
	readfs |= 1 << (*FP);
#else
	FD_SET((*FP),&readfs);
#endif
	if ((*FP) > biggest) biggest = (*FP);
      } else skip = 1;
    }

    if (skip) {
      tmp_ready = 0;
      timeout = &Timeout;
      Timeout.tv_sec = 0;
      Timeout.tv_usec = 0;
    } else if (0 != (timeout_key=timeout_callback(&atime))) {
      timeout = &Timeout;
      Timeout.tv_sec = atime / 10L;
      Timeout.tv_usec = ( atime % 10L) * 100000L;
    } else timeout = NULL;

    if (DEBUG_FILIO || DEBUG_TERM || DEBUG_BUFFER) {
      print_debug("Doing select: max fd=%d, %s (skip=%d)",
		  biggest, timeout ? "timeout" : "no timeout",
		  skip);
    }

#ifdef _HPUX_SOURCE
    res = select(biggest+1,&readfs,&writefs,(int *)NULL,
		     timeout);
#else
    res = select(biggest+1,&readfs,&writefs,(fd_set *)NULL,
		     timeout);
#endif

    if (-1 == res) {
      int code = errno;
      if (DEBUG_FILIO) {
	print_debug("Got errno=%d from select",
		    code);
      }
      if (EINTR == code) {
	tmp_ready = 0;
	continue;
      }
      if (have_file && ENXIO == code && !(*NOT_SELECTABLE)) {
	not_selectable(FP);
	tmp_ready = 0;
	continue;
      }
      print_error(code,
		  "(%s) Error when waiting response of terminal from /dev/tty",
		  prog);
      quitflag = 1;
      *ready = 1;
      continue;
    }

    if (0 == res && !skip) 
      (*timeout_handler)(timeout_key);

    if (
#ifdef _HPUX_SOURCE
	readfs & (1 << terminal)
#else
	FD_ISSET(terminal,&readfs)
#endif
	)
      {	
	CHAR input_buffer[INPUT_BUFFER_SIZE+1];
	int len = read_from_terminal(input_buffer);
	if (len > 0) parse_input(input_buffer, len); 
	was_input = 1;
      }
    
    cancel = 0;

    for (i = 0; command_len > i;) {
      CHAR *name;
      if (DEBUG_TERM) {
	CHAR *R1=printable_command(i);
	print_debug("Command(%d): %s",i,R1);
	FREE(R1);
      }
      if (do_escape(i)) {
	rem_command(i);
	if (DEBUG_TERM) {
	  print_debug("Command(%d) - handled by module control\n"
		      "            : REMOVED from buffer",i);
	}
      } else if(call_parser) {
	if (cancel || do_keyboard(i,&cancel)) {
	  if (DEBUG_TERM) {
	    print_debug("Command(%d) - handled (or canceled) "
			"by module keyboard",i);
	  }
	} else if (0 != (name = key_name(i))) {
	  char *buffer1 = "Key ", *buffer3 = " haven't action";
	  int bl1 = strlen(buffer1), bl3 = strlen(buffer3);
	  int bl2 = strlen(Cs(name));
	  CHAR_IDX buffer[50];
	  map_input(MAP_ASCII,bl1,rCs(buffer1),buffer);
	  map_input(MAP_ASCII,bl2,name,buffer+bl1);
	  map_input(MAP_ASCII,bl3,rCs(buffer3),buffer+bl1+bl2);
	  message_to_error(bl1+bl2+bl3,buffer);
	} else {
	  CHAR *R1=printable_sequence(i);
	  print_notify("Unexpected escape sequence: %s",R1);
	  free(R1);
	}
	rem_command(i);
	if (DEBUG_TERM) {
	  print_debug("Command(%d) : REMOVED form buffer",i);
	}
      } else {
	if (DEBUG_TERM) {
	  print_debug("Command(%d) : SKIPPED",i);
	}
	i++;
      }
    }

    (*callback_after)();


    if (have_file && -1 != (*FP) && !(*EOF_IN_FILE) &&     
	((*NOT_SELECTABLE) ||
#ifdef _HPUX_SOURCE
	 (readfs & (1 << (*FP)))
#else
	 FD_ISSET((*FP),&readfs)
#endif
	))
      {
	(*file_handler)();
      }

    if (
#ifdef _HPUX_SOURCE
	writefs & (1 << terminal)
#else
	FD_ISSET(terminal,&writefs)
#endif
	)
      { 
	int val = need_line_flush();
	tmp_ready = 0;
	if (was_input && 2 != val) {
	  if (DEBUG_BUFFER || DEBUG_LBUFFER) 
	    print_debug("BUFFER: was_input=%d, "
			"need_line_flush()=%d - no flush",
			was_input,val);
	} else {
	  if (DEBUG_BUFFER || DEBUG_LBUFFER) 
	    print_debug("BUFFER: was_input=%d, "
			"need_line_flush()=%d - flushing...",
			was_input,val);
	  if (!flush_some_lines())
	    flush_buffer(0);
	}
      }
    
  }

}

static int init_ready = 0;

static void init_do_timeout(int key) {
  do_timeout();
}

static int init_callback_before(void) {
  int skip = 0;
  int tail_ptr;

  if (got_from_TAIL(&tail_ptr)) {
    print_errors(tail_ptr);
    skip = 1; 
  }

  if (pager_state || quitflag) {
    init_ready = 1;
    skip = 1;
  }
  
  return skip;
}

static void init_callback_after(void) {

  if (pager_state || quitflag) {
    init_ready = 1;
  }
}

int init_schedule(void) {
  init_ready = 0;

  real_schedule(NULL,NULL,NULL,
		NULL,
		&ask_timeout,&init_do_timeout,
		&init_ready,&init_callback_before,
		&init_callback_after,0);

  return pager_state;

}
